mod nav_bar;
mod song_edit_window;


use std::collections::HashSet;

use egui::{Button, Color32, Label, RichText, Sense};
use egui_extras::{Column, TableBuilder};
use song_edit_window::{GuiError, GuiImportPlaylist, GuiNewSong};

use crate::{config::{Config, ConfigWrapper}, downloader::Downloader, manifest::{song::{Song, SongType}, Manifest}};

use self::song_edit_window::GuiSongEditor;

#[derive(Debug, Default)]
pub struct Gui {
    manifest: Manifest,
    song_edit_w: GuiSongEditor,
    new_song_w: GuiNewSong,
    import_playlist_w: GuiImportPlaylist,
    error_w: GuiError,
    filter: String,
    downloader: Downloader,
    cfg: ConfigWrapper,
    downloading: bool,
}

impl Gui {
    fn new(_: &eframe::CreationContext<'_>, manifest: Manifest, downloader: Downloader, cfg: ConfigWrapper) -> Self {
        Self {
            manifest,
            downloader,
            cfg,
            ..Default::default()
        }
    }

    pub fn start(manifest: Manifest, downloader: Downloader, cfg: ConfigWrapper) -> anyhow::Result<()> {
        let native_options = eframe::NativeOptions {
            viewport: egui::ViewportBuilder::default()
                .with_inner_size([400.0, 300.0])
                .with_min_inner_size([300.0, 220.0]),
                // .with_icon(
                //     // NOTE: Adding an icon is optional
                //     eframe::icon_data::from_png_bytes(&include_bytes!("../assets/icon-256.png")[..])
                //         .expect("Failed to load icon"),
                // ),
            ..Default::default()
        };

        if let Err(e) = eframe::run_native(
            "eframe template",
            native_options,
            Box::new(|cc| Box::new(Gui::new(cc, manifest, downloader, cfg))),
        ) {
            log::error!("Failed to create window: {e}");
        };

        Ok(())
    }
    pub fn throw_error(&mut self, text: impl ToString) {
        self.error_w.is_open = true;
        self.error_w.text = text.to_string();
    }
}

impl eframe::App for Gui {
    fn update(&mut self, ctx: &egui::Context, frame: &mut eframe::Frame) {
        self.draw_nav(ctx, frame);
        self.draw_song_edit_window(ctx, frame);
        self.draw_new_song_window(ctx, frame);
        self.draw_import_playlist_window(ctx, frame);
        self.draw_error_window(ctx, frame);

        egui::CentralPanel::default().show(ctx, |ui| {
            // The central panel the region left after adding TopPanel's and SidePanel's
            //ui.heading(format!("Songs ({})", self.manifest.get_song_count()));
            
            let fltr_by;
            let filter_clean;
            if self.filter.starts_with("playlist:") {
                fltr_by = "playlist";
                filter_clean = self.filter.strip_prefix("playlist:").unwrap_or("").to_string().to_lowercase();
            } else if self.filter.starts_with("source:") {
                fltr_by = "source";
                filter_clean = self.filter.strip_prefix("source:").unwrap_or("").to_string().to_lowercase();
            } else if self.filter.starts_with("url:") {
                fltr_by = "url";
                filter_clean = self.filter.strip_prefix("url:").unwrap_or("").to_string();
            } else {
                fltr_by = "";
                filter_clean = self.filter.clone();
            }

            ui.vertical(|ui| {
                ui.horizontal(|ui| {
                    ui.colored_label(Color32::from_hex("#4444aa").unwrap(), "Filter: ");
                    ui.text_edit_singleline(&mut self.filter);
                });  
            });
            
            ui.vertical(|ui| {
                let available_height = ui.available_height();
                let table = TableBuilder::new(ui)
                    .striped(true)
                    .cell_layout(egui::Layout::left_to_right(egui::Align::Center))
                    .resizable(true)
                    //.column(Column::auto())
                    .column(Column::auto())
                    //.column(
                    //    Column::remainder()
                    //        .at_least(40.0)
                    //        .clip(true)
                    //        .resizable(true),
                    //)
                    .column(Column::auto())
                    .column(Column::remainder())
                    //.column(Column::remainder())
                    .min_scrolled_height(0.0)
                    .max_scroll_height(available_height)
                    .sense(egui::Sense::click());
                
                let playlists = self.manifest.get_playlists().clone();

                let songs = {
                    let mut songs = Vec::new();
                    for (pname, p) in playlists {
                        for (sname, s) in p {
                            songs.push((pname.clone(), sname, s))
                        }
                    }
                    songs
                };

                table.header(20.0, |mut header| {
                    // header.col(|_|{});
                    header.col(|ui| {
                        ui.strong("Playlist");
                    }); 
                    header.col(|ui| { 
                        ui.strong("Source");
                    });
                    header.col(|ui| {
                        ui.strong("Name");
                    });
                }).body(|mut body| {
                    for (pname, sname, s) in songs {    
                        if fltr_by == "playlist" && !filter_clean.is_empty() {
                            if !pname.to_lowercase().contains(&filter_clean) {
                                continue;
                            }
                        } else if fltr_by == "type" && !filter_clean.is_empty(){
                            if !s.get_type().to_string().to_lowercase().contains(&filter_clean) {
                                continue;
                            }
                        } else if fltr_by == "url" && !filter_clean.is_empty(){
                            if !s.get_url_str().contains(&filter_clean) {
                                continue;
                            }
                        } else if !filter_clean.is_empty() {
                            if !sname.to_lowercase().contains(&filter_clean) {
                                continue;
                            }
                        }
                        body.row(18.0, |mut row| {
                             
                            row.col(|ui| {
                                ui.label(pname.clone())
                                    .context_menu(|ui| context_menu(self, ui, &pname, &sname, &s));
                            });
                            row.col(|ui| {
                                let color =
                                match s.get_type() {
                                    SongType::Youtube => Color32::from_hex("#FF0000").unwrap(),
                                    SongType::Spotify => Color32::from_hex("#1db954").unwrap(),
                                    SongType::Soundcloud => Color32::from_hex("#F26F23").unwrap()
                                };

                                ui.colored_label(color, s.get_type().to_string())
                                    .context_menu(|ui| context_menu(self, ui, &pname, &sname, &s));
                            });
                            row.col(|ui| {
                                ui.hyperlink_to(sname.clone(), s.get_url_str())
                                    .context_menu(|ui| context_menu(self, ui, &pname, &sname, &s));
                            });

                            row.response()
                                .context_menu(|ui| context_menu(self, ui, &pname, &sname, &s));

                            fn context_menu(this: &mut Gui, ui: &mut egui::Ui, pname: &String, sname: &String, song: &Song) {
                                if ui.button("Edit").clicked() {
                                    this.song_edit_w.song = (
                                        pname.clone(),
                                        sname.clone(),
                                    );
                                    this.song_edit_w.is_open = true;
                                    this.song_edit_w.ed_name = sname.clone();
                                    this.song_edit_w.ed_url = song.get_url_str().clone();
                                    ui.close_menu()
                                }

                                if ui.button("Download").clicked() {
                                    if let Err(e) = this.downloader.download_song_nb(&this.cfg, pname, sname, song, this.manifest.get_format()) {
                                        log::error!("{e}");
                                        this.throw_error(format!("Failed to download song {sname}: {e}"));
                                    }
                                    ui.close_menu()
                                }

                                if ui.button("Open Source").clicked() {
                                    if let Err(e) = open::that(song.get_url_str()) {
                                        log::error!("{e}");
                                        this.throw_error(format!("Failed to open song source: {e}"));
                                    }
                                    ui.close_menu()
                                }
                                if ui.button("Play").clicked() {
                                    let p = crate::util::get_song_path(pname, sname, this.manifest.get_format());
                                    
                                    if !p.exists() {
                                       this.throw_error(format!("Song does not exist on disk")); 
                                    } else if let Err(e) = open::that(p) {
                                        log::error!("{e}");
                                        this.throw_error(format!("Failed to play song: {e}"));
                                    }
                                    ui.close_menu()
                                }
                                if ui.button("Delete from disk").clicked() {
                                    let p = crate::util::get_song_path(pname, sname, this.manifest.get_format());
                                    if p.exists() {
                                        if let Err(e) = std::fs::remove_file(p) {
                                            this.throw_error(format!("Failed to delete file: {e}"));
                                        }
                                    }
                                    ui.close_menu();
                                }
                                if ui.button(RichText::new("Delete").color(Color32::RED)).clicked() {
                                    this.throw_error("TODO");
                                    ui.close_menu()
                                }
                            }
                        })
                    }
                });
            });
            ui.separator();


            ui.with_layout(egui::Layout::bottom_up(egui::Align::LEFT), |ui| {
                egui::warn_if_debug_build(ui);
            });
        });
    }
}
