
use std::sync::Arc;

use egui::{text::{LayoutJob, TextWrapping}, Color32, Label, RichText, Style, TextBuffer, TextFormat, TextStyle};


use crate::manifest::{playlist::Playlist, song::{Song, SongType}};

use super::Gui;

#[derive(Debug, Default)]
pub struct GuiSongEditor {
    pub is_open: bool,
    pub song: (String, String),
    pub ed_url: String,
    pub ed_name: String,
}

#[derive(Debug, Default)]
pub struct GuiNewSong {
    pub is_open: bool,
    ed_type: SongType,
    ed_name: String,
    ed_playlist: Option<String>,
    ed_url: String,
}

#[derive(Debug, Default)]
pub struct GuiImportPlaylist {
    pub is_open: bool,
    ed_name: String,
    ed_url: String,
}

#[derive(Debug, Default)]
pub struct GuiError {
    pub is_open: bool,
    pub text: String,
}

impl Gui {
    pub fn draw_song_edit_window(&mut self, ctx: &egui::Context, _: &mut eframe::Frame) {
        let mut save = false;
        
        let (playlist, song_name) = self.song_edit_w.song.clone();

        let Some(song) = self.manifest.get_song(&playlist, &song_name) else {
            return;
        };
        let song = song.clone();
        
        egui::Window::new("Song editor")
            .open(&mut self.song_edit_w.is_open)
            .show(ctx, 
            |ui| {

            ui.horizontal(|ui| {
                ui.spacing_mut().item_spacing.x = 0.0;
                ui.label("[");
                ui.hyperlink_to("link", song.get_url().unwrap());
                ui.label("] ");
                ui.colored_label(Color32::LIGHT_BLUE, &playlist);
                ui.label(": ");
                ui.label(&song_name)
            });

            ui.horizontal(|ui| {
                ui.label("Type: ");
                ui.label(&song.get_type().to_string());
            });

            ui.horizontal(|ui| {
                ui.label("Name: ");
                ui.text_edit_singleline(&mut self.song_edit_w.ed_name);
            });
            ui.horizontal(|ui| {
                ui.label("Url: ");
                ui.text_edit_singleline(&mut self.song_edit_w.ed_url);
            });

            if ui.button("Save").clicked() {
                save = true;
            }
        });

        if save {
            {
                let Some(song) = self.manifest.get_song_mut(&playlist, &song_name) else {
                    return;
                };
                
                *song.get_url_str_mut() = self.song_edit_w.ed_url.clone();
            }

            let Some(playlist) = self.manifest.get_playlist_mut(&playlist) else {
                return;
            };


            playlist.remove_song(&song_name);
            playlist.add_song(self.song_edit_w.ed_name.clone(), song);
            self.song_edit_w.is_open = false;
            let _ = self.manifest.save(None);
        }
    }
    
    pub fn draw_new_song_window(&mut self, ctx: &egui::Context, _: &mut eframe::Frame) {
        let mut save = false;
        egui::Window::new("New song")
            .open(&mut self.new_song_w.is_open)
            .show(ctx, |ui| {
                ui.horizontal(|ui| {
                    ui.label("Type: ");
                    egui::ComboBox::from_id_source("new_song_window_type")
                        .selected_text(format!("{:?}", self.new_song_w.ed_type))
                        .show_ui(ui, |ui| {
                            ui.selectable_value(&mut self.new_song_w.ed_type, SongType::Youtube, "Youtube");
                            ui.selectable_value(&mut self.new_song_w.ed_type, SongType::Spotify, "Spotify");
                            ui.selectable_value(&mut self.new_song_w.ed_type, SongType::Soundcloud, "Soundcloud");
                        }
                    );    
                });

                ui.horizontal(|ui| {
                    ui.label("Name: ");
                    ui.text_edit_singleline(&mut self.new_song_w.ed_name);
                });
                ui.horizontal(|ui| {
                    ui.label("Playlist: ");
                    egui::ComboBox::from_id_source("new_song_window_playlist")
                        .selected_text(format!("{}", self.new_song_w.ed_playlist.clone().unwrap_or("".to_string())))
                        .show_ui(ui, |ui| {
                            for p in self.manifest.get_playlists().keys() {
                                ui.selectable_value(&mut self.new_song_w.ed_playlist, Option::Some(p.clone()), p.as_str());
                            }
                        }
                    );    
                });
                ui.horizontal(|ui| {
                    ui.label("Url: ");
                    ui.text_edit_singleline(&mut self.new_song_w.ed_url);
                });

                if ui.button("Save").clicked() {
                    save = true;
                }
        });

        if save {
            let Some(playlist) = self.manifest.get_playlist_mut(&self.new_song_w.ed_playlist.clone().unwrap())  else {
                panic!("couldnt find playlist from a preset playlist list????????????");
            };

            playlist.add_song(
                self.new_song_w.ed_name.clone(),
                Song::from_url_str(self.new_song_w.ed_url.clone()).unwrap().set_type(self.new_song_w.ed_type.clone()).clone()
            );


            let _ = self.manifest.save(None);
            self.new_song_w.is_open = false;
        }
    }

    pub fn draw_import_playlist_window(&mut self, ctx: &egui::Context, _: &mut eframe::Frame) {
        let mut save = false;
        egui::Window::new("Import Playlist")
            .open(&mut self.import_playlist_w.is_open)
            .show(ctx, |ui| {
                ui.horizontal(|ui| {
                    ui.label("Type: Youtube");
                });

                ui.horizontal(|ui| {
                    ui.label("Name: ");
                    ui.text_edit_singleline(&mut self.import_playlist_w.ed_name);
                });
                ui.horizontal(|ui| {
                    ui.label("Url: ");
                    ui.text_edit_singleline(&mut self.import_playlist_w.ed_url);
                });

                if ui.button("Import").clicked() {
                    save = true;
                }
        });

        if save {
            let name = self.import_playlist_w.ed_name.clone();
            let url = self.import_playlist_w.ed_url.clone();

            if self.manifest.get_playlist(&name).is_some() {
                log::error!("Playlist {name} already exists");
                self.throw_error(format!("Playlist {name} already exists"));
            }

            let songs = self.downloader.download_playlist_nb(&self.cfg, &url, &name, &self.manifest.get_format()).unwrap();
            self.manifest.add_playlist(name.clone());

            let playlist = self.manifest.get_playlist_mut(&name).expect("Unreachable");
            
            for (sname, song) in songs {
                log::info!("Added: {sname}");
                playlist.add_song(sname, song);
            }
            let _ = self.manifest.save(None);
            self.import_playlist_w.is_open = false;
        }
    }

    pub fn draw_error_window(&mut self, ctx: &egui::Context, _: &mut eframe::Frame) {
        egui::Window::new("ERROR!!!! D:")
            .open(&mut self.error_w.is_open)
            .show(ctx, |ui| {
                ui.vertical(|ui| {
                    ui.label(RichText::new("Error:").size(30.0).color(Color32::RED));
                    ui.horizontal(|ui| {
                        ui.add(Label::new(self.error_w.text.clone()).wrap(true));
                    })
                })
        });
        
    }

    
}
